// ***************************************************************************************
// File: javaint.cpp
// Description: interface from nvs to java
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: the real java interface is implemented in a dedicated library
//           named nvjava. This file is the layer that interfaces to this
//           library.
// ***************************************************************************************


#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvJava

NvJava::NvJava()
{
	hLibJava = NULL;
	JavaLibInit = NULL;
	JavaLibFree = NULL;
	JavaLibRunMethod = NULL;
	JavaLibDetachThread = NULL;
	NumInstances = 0;
	m_bAvailable = false;
}

NvJava::~NvJava()
{
	// Unload the java interface library if loaded
	FreeJavaLib();
}


bool NvJava::LoadJavaLib(const char *NirvaDir, int MinMemSize, int MaxMemSize, bool UseLocking, const char* JavaHomeDir, const char* JvmPath, const char* JvmParam)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectJava);
	if(hLibJava != NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectJava);
		return true;
	}

	NumInstances = 0;

	if(MinMemSize < 4)
		MinMemSize = 4;
	if(MaxMemSize < MinMemSize)
		MaxMemSize = MinMemSize;

	// Load the library here
	string JavaLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(JavaLibPath);
	JavaLibPath += "Bin";
	AddDirSuffix(JavaLibPath);
	JavaLibPath += "nvjava";
#ifdef WIN32
#ifdef NV_DEBUG
	JavaLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(JavaLibPath);
	JavaLibPath += "Bin";
	AddDirSuffix(JavaLibPath);
	JavaLibPath += "nvjavad";
	//JavaLibPath = GetMain()->GetNirvaDir();
	//AddDirSuffix(JavaLibPath);
	//JavaLibPath += "Sources";
	//AddDirSuffix(JavaLibPath);
	//JavaLibPath += "nvjava";
	//AddDirSuffix(JavaLibPath);
	//JavaLibPath += "Debug";
	//AddDirSuffix(JavaLibPath);
	//JavaLibPath += "nvjava";
#endif
	JavaLibPath += ".dll";
#endif
#ifdef LINUX
	JavaLibPath += ".so";
#endif
#ifdef HPUX
	JavaLibPath += ".sl";
#endif
#ifdef HPUXI
	JavaLibPath += ".so";
#endif
#ifdef SOLARIS
	JavaLibPath += ".so";
#endif
#ifdef AIX
	JavaLibPath += ".a";
#endif

	hLibJava = MyLoadLibrary(JavaLibPath.c_str());
	if(hLibJava == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectJava);
		return false;
	}

	JavaLibInit = (JAVAINIT)MyGetLibFunction(hLibJava, "NvInitJava");
	JavaLibFree = (JAVAFREE)MyGetLibFunction(hLibJava, "NvFreeJava");
	JavaLibRunMethod = (JAVARUNMETHOD)MyGetLibFunction(hLibJava, "NvRunMethod");
	JavaLibDetachThread = (JAVADETACHTHREAD)MyGetLibFunction(hLibJava, "NvDetachThread");
	if((JavaLibInit == NULL) || (JavaLibFree == NULL) || (JavaLibRunMethod == NULL) || (JavaLibDetachThread == NULL))
	{
		MyFreeLibrary(hLibJava);
		JavaLibInit = NULL;
		JavaLibFree = NULL;
		JavaLibRunMethod = NULL;
		JavaLibDetachThread = NULL;
		hLibJava = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectJava);
		return false;
	}

	char sMinSize[50];
	char sMaxSize[50];
	sprintf(sMinSize, "-Xms%dM", MinMemSize);
	sprintf(sMaxSize, "-Xmx%dM", MaxMemSize);
	if((JavaLibInit)(NirvaDir, sMinSize, sMaxSize, JavaHomeDir, JvmPath, JvmParam) == 0)
	{
		MyFreeLibrary(hLibJava);
		JavaLibInit = NULL;
		JavaLibFree = NULL;
		JavaLibRunMethod = NULL;
		JavaLibDetachThread = NULL;
		hLibJava = NULL;
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectJava);
		return false;
	}


	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectJava);

	return true;
}



void NvJava::FreeJavaLib(bool UseLocking)
{
	if(UseLocking)
		NV_LOCKLOCK(&ProtectJava);
	NumInstances = 0;
	if(hLibJava == NULL)
	{
		if(UseLocking)
			NV_UNLOCKLOCK(&ProtectJava);
		return;
	}

	(JavaLibFree)();
	MyFreeLibrary(hLibJava);
	JavaLibInit = NULL;
	JavaLibFree = NULL;
	JavaLibRunMethod = NULL;
	JavaLibDetachThread = NULL;
	hLibJava = NULL;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectJava);
}



void NvJava::CleanJavaLib(const char *NirvaDir, int MinMemSize, int MaxMemSize, const char* JavaHomeDir, const char* JvmPath, const char* JvmParam)
{
	NV_LOCKLOCK(&ProtectJava);
	if(NumInstances)
	{
		// No cleaning while in use
		NV_UNLOCKLOCK(&ProtectJava);
		return;
	}
	if(!m_bAvailable)
	{
		// No cleaning if not available
		NV_UNLOCKLOCK(&ProtectJava);
		return;
	}
	FreeJavaLib(false);
	LoadJavaLib(NirvaDir, MinMemSize, MaxMemSize, false, JavaHomeDir, JvmPath, JvmParam);

	NV_UNLOCKLOCK(&ProtectJava);
}



bool NvJava::IsJavaLibLoaded(bool UseLocking)
{
	bool RetVal = true;
	if(UseLocking)
		NV_LOCKLOCK(&ProtectJava);
	if(hLibJava == NULL)
		RetVal = false;
	if(UseLocking)
		NV_UNLOCKLOCK(&ProtectJava);
	return RetVal;
}


void NvJava::DetachJavaThread()
{
	if(IsJavaLibLoaded())
		(JavaLibDetachThread)();
}


// Run a java method from .
int NvJava::RunMethod(const char* Method, void *pCommand)
{
	NV_LOCKLOCK(&ProtectJava);
	if(!IsJavaLibLoaded(false))
	{
		NV_UNLOCKLOCK(&ProtectJava);
		return 0;
	}
	NumInstances++;
	NV_UNLOCKLOCK(&ProtectJava);
	int RetCode = 1;
	try
	{
		RetCode = (JavaLibRunMethod)(Method, pCommand);
	}
	catch(...)
	{
		RetCode = 0;
	}
	NV_LOCKLOCK(&ProtectJava);
	if(NumInstances > 0)
		NumInstances--;
	NV_UNLOCKLOCK(&ProtectJava);
	return RetCode;

}


int NvJava::ThreadRunMethod(const char* Method, void *pCommand, bool SameThread)
{
	NvJavaMethod *JavaMethod = NVNEW NvJavaMethod;
	if(JavaMethod == NULL)
		return 0;

	int RetCode = 1;
	try
	{
		RetCode = JavaMethod->RunMethod(Method, pCommand, SameThread);
	}
	catch(...)
	{
		RetCode = 0;
	}

	delete JavaMethod;

	return RetCode;
}


NvJavaMethod::NvJavaMethod()
{
	b_Initialized = false;
}


int NvJavaMethod::Main()
{
	return GetMain()->JavaInterface.RunMethod(MethodPath, pCommandHere);
}


// called before the Main function (see class NvTask)
bool NvJavaMethod::OnInit()
{
	NvTask::OnInit();	// call the base class

	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_JAVA);
	
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type JAVA", NV_LOG_INFO, 1, sThreadId, "system");

	b_Initialized = true;

	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type JAVA", NV_LOG_INFO, 1, sThreadId, "system");
	return true;
}



// called after the Main function (see class NvTask)
void NvJavaMethod::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type JAVA", NV_LOG_INFO, 1, sThreadId, "system");

	// Detach thread from java interface if not yet made
	if(!m_bSameThread)
		GetMain()->JavaInterface.DetachJavaThread();

	// to call before leaving
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type JAVA", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}



bool NvJavaMethod::IsJavaAvailable()
{
	return GetMain()->JavaInterface.IsJavaLibLoaded();
}


// Run a java method
// The java method has the following format: methodclasspath:methodname
// Typically, the methodclasspath will the a class file in the procedure directory
// The method name is the name of a static method embedded in the given class file
// If not provided, the function will use main or Main
int NvJavaMethod::RunMethod(const char* Method, void *pCommand, bool SameThread)
{
	if(!IsJavaAvailable())
		return 0;
	MethodPath = Method;
	pCommandHere = pCommand;

	if(!Run(SameThread))
		return 0;

	if(!SameThread)
	{
		// Wait for end of thread but let other threads do their work
		int CheckEvery = 1;
		while(IsThreadRunning())
		{
			SLEEP(CheckEvery);

			// We first do the check very often but less and less while the time is growing
			CheckEvery += (CheckEvery/10)+1;
			if(CheckEvery > 10000)		// When the check time is 10 seconds, we check every 10 seconds
				CheckEvery = 10000;
		}
	}

	if(!m_iExitCode)
		return 0;
	return 1;
}



