// ***************************************************************************************
// File: perlint.cpp
// Description: interface from nvs to perl
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: the real perl interface is implemented in a dedicated library
//           named nvperl. This file is the layer that interfaces to this
//           library.
// ***************************************************************************************

#include "nvs.h"

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER


using namespace std;



/////////////////////////////////////////////////////////////////////////////
// NvPerl

NvPerl::NvPerl()
{
	hLibPerl = NULL;
	PerlLibGetInterpretor = NULL;
	PerlLibFreeInterpretor = NULL;
	PerlLibRunScript = NULL;
	PerlLibRunScriptMemory = NULL;
	PerlInitLib = NULL;
	PerlExitLib = NULL;
	NumInstances = 0;
}

NvPerl::~NvPerl()
{
	// Unload the perl interface library if loaded
	FreePerlLib();
}


bool NvPerl::LoadPerlLib()
{
	NV_LOCKLOCK(&ProtectPerl);
	if(hLibPerl != NULL)
	{
		NV_UNLOCKLOCK(&ProtectPerl);
		return true;
	}
	// Load the library here
	string PerlLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(PerlLibPath);
	PerlLibPath += "Bin";
	AddDirSuffix(PerlLibPath);
	PerlLibPath += "nvperl";
#ifdef WIN32
#ifdef NV_DEBUG
	PerlLibPath = GetMain()->GetNirvaDir();
	AddDirSuffix(PerlLibPath);
	PerlLibPath += "Bin";
	AddDirSuffix(PerlLibPath);
	PerlLibPath += "nvperld";
	//PerlLibPath = GetMain()->GetNirvaDir();
	//AddDirSuffix(PerlLibPath);
	//PerlLibPath += "Sources";
	//AddDirSuffix(PerlLibPath);
	//PerlLibPath += "nvperl";
	//AddDirSuffix(PerlLibPath);
	//PerlLibPath += "Debug";
	//AddDirSuffix(PerlLibPath);
	//PerlLibPath += "nvperl";
#endif
	PerlLibPath += ".dll";
#endif
#ifdef LINUX
	PerlLibPath += ".so";
#endif
#ifdef HPUX
	PerlLibPath += ".sl";
#endif
#ifdef HPUXI
	PerlLibPath += ".so";
#endif
#ifdef SOLARIS
	PerlLibPath += ".so";
#endif
#ifdef AIX
	PerlLibPath += ".a";
#endif

	hLibPerl = MyLoadLibrary(PerlLibPath.c_str(), true);
	if(hLibPerl == NULL)
	{
		NV_UNLOCKLOCK(&ProtectPerl);
		return false;
	}

	PerlLibGetInterpretor = (PERLGETINTERPRETOR)MyGetLibFunction(hLibPerl, "NvGetInterpretor");
	PerlLibFreeInterpretor = (PERLFREEINTERPRETOR)MyGetLibFunction(hLibPerl, "NvFreeInterpretor");
	PerlLibRunScript = (PERLRUNSCRIPT)MyGetLibFunction(hLibPerl, "NvInterpretorRunScript");
	PerlLibRunScriptMemory = (PERLRUNSCRIPTMEMORY)MyGetLibFunction(hLibPerl, "NvInterpretorRunScriptMemory");
	if((PerlLibGetInterpretor == NULL) || (PerlLibFreeInterpretor == NULL) || (PerlLibRunScript == NULL))
	{
		MyFreeLibrary(hLibPerl);
		PerlLibGetInterpretor = NULL;
		PerlLibFreeInterpretor = NULL;
		PerlLibRunScript = NULL;
		PerlLibRunScriptMemory = NULL;
		hLibPerl = NULL;
		NV_UNLOCKLOCK(&ProtectPerl);
		return false;
	}

	PerlInitLib = (INITPERLLIB)MyGetLibFunction(hLibPerl, "InitPerlLib");
	PerlExitLib = (EXITPERLLIB)MyGetLibFunction(hLibPerl, "ExitPerlLib");
	if(PerlInitLib != NULL)
	{
		if(!(PerlInitLib)(GetMain()->GetNirvaDir().c_str()))
		{
			MyFreeLibrary(hLibPerl);
			PerlLibGetInterpretor = NULL;
			PerlLibFreeInterpretor = NULL;
			PerlLibRunScript = NULL;
			PerlLibRunScriptMemory = NULL;
			hLibPerl = NULL;
			NV_UNLOCKLOCK(&ProtectPerl);
			return false;
		}
	}

	NV_UNLOCKLOCK(&ProtectPerl);

	return true;
}


void NvPerl::AutoFree()
{
	NV_LOCKLOCK(&ProtectPerl);
	if(hLibPerl == NULL)
	{
		NV_UNLOCKLOCK(&ProtectPerl);
		return;
	}
	if(NumInstances > 0)
	{
		NV_UNLOCKLOCK(&ProtectPerl);
		return;
	}

	try
	{
		if(PerlExitLib != NULL)
			(PerlExitLib)();
		MyFreeLibrary(hLibPerl);
	}
	catch(...)
	{
	}
	hLibPerl = NULL;
	PerlLibGetInterpretor = NULL;
	PerlLibFreeInterpretor = NULL;
	PerlLibRunScript = NULL;
	PerlLibRunScriptMemory = NULL;
	PerlInitLib = NULL;
	PerlExitLib = NULL;
	NV_UNLOCKLOCK(&ProtectPerl);
}

void NvPerl::FreePerlLib()
{
	NV_LOCKLOCK(&ProtectPerl);
	if(hLibPerl == NULL)
	{
		NV_UNLOCKLOCK(&ProtectPerl);
		return;
	}
	
	try
	{
		if(PerlExitLib != NULL)
			(PerlExitLib)();
		MyFreeLibrary(hLibPerl);
	}
	catch(...)
	{
	}
	hLibPerl = NULL;
	PerlLibGetInterpretor = NULL;
	PerlLibFreeInterpretor = NULL;
	PerlLibRunScript = NULL;
	PerlLibRunScriptMemory = NULL;
	PerlInitLib = NULL;
	PerlExitLib = NULL;
	NV_UNLOCKLOCK(&ProtectPerl);
}

bool NvPerl::IsPerlLibLoaded()
{
	bool RetVal = true;
	NV_LOCKLOCK(&ProtectPerl);
	if(hLibPerl == NULL)
		RetVal = false;
	NV_UNLOCKLOCK(&ProtectPerl);
	return RetVal;
}

NVPERLINTERPRETOR NvPerl::PerlGetInterpretor()
{
	NV_LOCKLOCK(&ProtectPerl);
	NumInstances++;
	NV_UNLOCKLOCK(&ProtectPerl);
	
	NVPERLINTERPRETOR Perli = NULL;	
	
	if(LoadPerlLib())
		Perli = (PerlLibGetInterpretor)();
	if(Perli == NULL)
	{
		NV_LOCKLOCK(&ProtectPerl);
		if(NumInstances > 0)
			NumInstances--;
		NV_UNLOCKLOCK(&ProtectPerl);
	}
	return Perli;
}

void NvPerl::PerlFreeInterpretor(NVPERLINTERPRETOR Interpretor)
{
	if(!IsPerlLibLoaded())
		return;

	if(Interpretor == NULL)
		return;

	(PerlLibFreeInterpretor)(Interpretor);
	NV_LOCKLOCK(&ProtectPerl);
	if(NumInstances > 0)
		NumInstances--;
	NV_UNLOCKLOCK(&ProtectPerl);
}


int NvPerl::PerlRunScript(NVPERLINTERPRETOR Interpretor, const char* ScriptFileName, void *pCommand)
{
	if(!IsPerlLibLoaded())
		return 0;

	if(Interpretor == NULL)
		return 0;

	// For small files we use memory script
	GetFileLength(ScriptFileName);
	NvInt64 FileLength = GetFileLength(ScriptFileName);
	char *ScriptData = NULL;
	if(FileLength < 100000)
		ScriptData = GetScriptData(ScriptFileName);
	if((ScriptData != NULL) && (PerlLibRunScriptMemory == NULL))
	{
		if(ScriptData != NULL)
			free(ScriptData);
		return 0;
	}

	int RetVal = 1;
	int Debug = 0;
	if(GetMain()->GetDebug())
		Debug = 1;
	try
	{
		if(ScriptData != NULL)
			RetVal = (PerlLibRunScriptMemory)(Interpretor, ScriptData, pCommand, Debug);
		else
			RetVal = (PerlLibRunScript)(Interpretor, ScriptFileName, pCommand, Debug);
	}
	catch(...)
	{
		RetVal = 0;
	}
	if(ScriptData != NULL)
		free(ScriptData);
	return RetVal;
}

int NvPerl::PerlThreadRunScript(const char* ScriptFileName, void *pCommand, bool SameThread)
{
	NvPerlSript *PerlScript = NVNEW NvPerlSript;
	if(PerlScript == NULL)
		return 0;

	int RetCode = 1;
	try
	{
		RetCode = PerlScript->RunScript(ScriptFileName, pCommand, SameThread);
	}
	catch(...)
	{
		RetCode = 0;
	}

	delete PerlScript;
	return RetCode;
}


char* NvPerl::GetScriptData(const char* ScriptFileName)
{
	char *ScriptData = NULL;

	FILE* File =  Nvfopen(ScriptFileName, "rb");
	if(File == NULL)
		return NULL;

	char Header[50];
	Header[0] = '\0';
	freadline(Header, 49, File);
	string HeaderHere = Header;
	bool bIsPerlEncoded = false;
	string Ctrl = "";
	if(HeaderHere == "NIRVA PERL ENCODED FILE")
	{
		freadline(Header, 49, File);
		Ctrl = Header;
		if(Ctrl.length() == 8)
			bIsPerlEncoded = true;
	}
	if(!bIsPerlEncoded)
		fseek(File, 0, SEEK_SET);

	int BuffSize = 0;
	int CurPos = ftell(File);

	if(fseek(File, 0, SEEK_END) == 0)
		BuffSize = ftell(File) - CurPos; 

	fseek(File, CurPos, SEEK_SET);

	ScriptData = (char *)malloc(BuffSize + 1);
	if(ScriptData == NULL)
	{
		fclose(File);
		return NULL;
	}
	ScriptData[BuffSize] = '\0';
	if(fread(ScriptData, 1, BuffSize, File) != BuffSize)
	{
		fclose(File);
		free(ScriptData);
		ScriptData = NULL;
		return NULL;
	}
	fclose(File);

	if(bIsPerlEncoded)
	{
		string EncryptionKey = "p";
		EncryptionKey += "nri";
		EncryptionKey += "Aav";
		EncryptionKey += "m";
		EncryptionKey += "lerp";
		EncryptionKey += Ctrl;
		EncryptionKey = MD5_string(EncryptionKey).substr(0,10);
		XOR(ScriptData, BuffSize, EncryptionKey.c_str());
	}
	return ScriptData;
}




NvPerlSript::NvPerlSript()
{
	b_Initialized = false;
}


int NvPerlSript::Main()
{
	return GetMain()->PerlInterface.PerlRunScript(PerlInterpretor, ScriptFile, pCommandHere);
}


// called before the Main function (see class NvTask)
bool NvPerlSript::OnInit()
{
	NvTask::OnInit();	// call the base class
	
	if(!m_bSameThread)
		GetMain()->Threads.Add(THREAD_TYPE_PERL);

	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Init thread", "Type PERL", NV_LOG_INFO, 1, sThreadId, "system");

	PerlInterpretor = NULL;

	try
	{
		PerlInterpretor = GetMain()->PerlInterface.PerlGetInterpretor();
	}
	catch(...)
	{
		PerlInterpretor = NULL;
	}
	if(PerlInterpretor == NULL)
	{
		if(!m_bSameThread)
			GetMain()->Logs.Write("Threads", "Error init thread (Get interpretor)", "Type PERL", NV_LOG_ERROR, 1, sThreadId, "system");
		return false;
	}

	b_Initialized = true;
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread started", "Type PERL", NV_LOG_INFO, 1, sThreadId, "system");
	return true;
}



// called after the Main function (see class NvTask)
void NvPerlSript::OnCleanup()
{
	char sThreadId[50];
	sprintf(sThreadId, "%lu", NvGetCurrentThreadId());
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Cleanup thread", "Type PERL", NV_LOG_INFO, 1, sThreadId, "system");

	if(b_Initialized)
	{
		try
		{
			if(PerlInterpretor != NULL)
				GetMain()->PerlInterface.PerlFreeInterpretor(PerlInterpretor);

			if(!m_bSameThread)
				GetMain()->JavaInterface.DetachJavaThread();
		}
		catch(...)
		{
		}
	}

	// to call before leaving
	if(!m_bSameThread)
		GetMain()->Logs.Write("Threads", "Thread stopped", "Type PERL", NV_LOG_INFO, 1, sThreadId, "system");
	if(!m_bSameThread)
		GetMain()->Threads.Remove();
	NvTask::OnCleanup();	// call the base class
}



int NvPerlSript::RunScript(const char* ScriptFileName, void *pCommand, bool SameThread)
{
//	if(!GetMain()->PerlInterface.LoadPerlLib())
//		return 0;

	ScriptFile = ScriptFileName;
	pCommandHere = pCommand;

	if(!Run(SameThread))
		return 0;

	if(!SameThread)
	{
		// Wait for end of thread but let other threads do their work
		int CheckEvery = 1;
		while(IsThreadRunning())
		{
			SLEEP(CheckEvery);

			// We first do the check very often but less and less while the time is growing
			CheckEvery += (CheckEvery/10)+1;
			if(CheckEvery > 10000)		// When the check time is 10 seconds, we check every 10 seconds
				CheckEvery = 10000;
		}
	}

	if(!m_iExitCode)
		return 0;

	return 1;
}



