// ***************************************************************************************
// File: task.cpp
// Description: thread management functions and classes
// Author: Pierre MARC
// Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Nirva Systems
//
// This file is part of Nirva Application Platform.
// 
// Nirva Application Platform is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// Nirva Application Platform is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with Nirva Application Platform.  If not, see <http://www.gnu.org/licenses/>.
//
// ***************************************************************************************
// Comments: 
// ***************************************************************************************

#ifdef WIN32
#pragma warning( disable : 4996 )
#endif

#include "task.h"
#ifdef CHECK_DEADLOCKS
#include <map>
#endif

#ifdef NV_SERVER
	#ifdef WIN32
		#ifdef _DEBUG
			#include <stdlib.h>
			#include <crtdbg.h>
			#define NVNEW new( _NORMAL_BLOCK, __FILE__, __LINE__)
		#else // _DEBUG
			#define NVNEW new
		#endif // _DEBUG
	#else	//WIN32
		#define NVNEW new
	#endif	//WIN32
#else	//NV_SERVER
	#define NVNEW new
#endif	//NV_SERVER

using namespace std;


#ifndef WIN32
NvMutex ProtectUnixSignalFunc;
#endif

#ifdef HAVE_NV_STRING

Basic_NvString& Basic_NvString::append(const Basic_NvString& X)
{
	NvGrowIt();
	return (Basic_NvString&)(string::append(X));
}

Basic_NvString& Basic_NvString::append(const char *S)
{
	NvGrowIt();
	return (Basic_NvString&)(string::append(S));
}

Basic_NvString& Basic_NvString::append(char C)
{
	NvGrowIt();
	return (Basic_NvString&)(string::append(1, C));
}

Basic_NvString& Basic_NvString::append(const std::string& X)
{
	NvGrowIt();
	return (Basic_NvString&)(string::append(X));
}

// reserve string space by reallocating if necessary
void Basic_NvString::NvGrowIt(int NewSize)
{
	if(empty())
		return;
	if(NewSize > 0)
	{
		reserve(NewSize);
		return;
	}

	if((capacity() - size()) > size())
		return;

	int GrowBy = size();
	if(GrowBy > 1000)
		GrowBy = 1000;

	reserve(capacity() + GrowBy);
}
#endif



// convert a string to upper case
void MakeUpper(string& sString)
{
	for(int i=0; i<(int)sString.length(); i++)
		sString.at(i) = (char)(toupper(sString[i]));
}
// convert a string to lower case
void MakeLower(string& sString)
{
	for(int i=0; i<(int)sString.length(); i++)
		sString.at(i) = (char)(tolower(sString[i]));
}

// reserve string space by reallocating if necessary
void ReserveString(string& sString, int NewSize)
{
	if(!sString.size())
		return;
	if(NewSize > 0)
	{
		sString.reserve(NewSize);
		return;
	}

	if((sString.capacity() - sString.size()) > sString.size())
		return;

	int GrowBy = (int)sString.size();
	if(GrowBy > 1000)
		GrowBy = 1000;

	int NewStringSize = (int)sString.capacity() + GrowBy;
	sString.reserve(NewStringSize);
}

long NvGetCurrentThreadId()
{
	long CurrentThreadId = 0;
#ifdef WIN32
	CurrentThreadId = (long)GetCurrentThreadId();			// Thread handle for window
#else
	CurrentThreadId = (long)pthread_self();				// Thread handle for unix
#endif
	return CurrentThreadId;
}

// Deadlock checking specific code
#ifdef CHECK_DEADLOCKS
static bool CheckDeadlocksInitialized = false;
static bool bDisplayAll = false;
static char DeadlockBuffer[1024];
bool bCheckDeadlocks = false;
NvMutex ProtectCheckDeadLocks;
std::map<std::string,std::string> ActualLocks; // first string is the lock itself and second is the thread owner
string sDeadlockFileName = "nvdead.log";
void OnMutexTryLock(void * Mutex, const char *SourceFile, int SourceFileLine)
{
	ProtectCheckDeadLocks.Lock();
	string sSourceFile = GetSourceFileName(SourceFile);
	string sSourceFileLine = GetSourceFileLine(SourceFileLine);
	string sThreadId = GetCurrentThreadIdentifier();
	string sMutex = GetMutexString(Mutex);
	string sAction = "TRY LOCK";
	bool bDeadlockFound = false;
	if(ActualLocks.find(sMutex) != ActualLocks.end())
	{
		// The mutex is ever locked
		// Try to get the thread that locked it
		string sLockingThreadId = ActualLocks[sMutex];
		string sFileNameLine = "";
		string::size_type index;
		index = sLockingThreadId.find(';');
		if(index != string::npos)
		{
			sFileNameLine = sLockingThreadId.substr(index+1);
			sLockingThreadId = sLockingThreadId.substr(0, index);
		}
		if(sLockingThreadId == sThreadId)
		{
			bDeadlockFound = true;
			// This is a potential deadlock
			// A thread is trying to lock 2 times the same mutex
			sprintf(DeadlockBuffer, "*** Potential deadlock found for next mutex - Same thread tries to lock 2 times the same lock ***\n");
			WriteDeadlockFile();
			sprintf(DeadlockBuffer, "*** Origin mutex file and line is %s ***\n", sFileNameLine.c_str());
			WriteDeadlockFile();
		}
	}
	if(bDeadlockFound || bDisplayAll)
	{
		sprintf(DeadlockBuffer, "th:%s - mx:%s - %-15s - File = %-15s - Line = %s\n", sThreadId.c_str(), sMutex.c_str(), sAction.c_str(), sSourceFile.c_str(), sSourceFileLine.c_str());
		WriteDeadlockFile();
	}
	ProtectCheckDeadLocks.Unlock();
}
void OnMutexLock(void * Mutex, const char *SourceFile, int SourceFileLine)
{
	ProtectCheckDeadLocks.Lock();
	string sSourceFile = GetSourceFileName(SourceFile);
	string sSourceFileLine = GetSourceFileLine(SourceFileLine);
	string sThreadId = GetCurrentThreadIdentifier();
	string sMutex = GetMutexString(Mutex);
	string sAction = "LOCKING OK";
	bool bDeadlockFound = false;
	if(ActualLocks.find(sMutex) != ActualLocks.end())
	{
		// The mutex is ever in the locked list
		// Try to get the thread that locked it
		string sLockingThreadId = ActualLocks[sMutex];
		string sFileNameLine = "";
		string::size_type index;
		index = sLockingThreadId.find(';');
		if(index != string::npos)
		{
			sFileNameLine = sLockingThreadId.substr(index+1);
			sLockingThreadId = sLockingThreadId.substr(0, index);
		}
		if(sLockingThreadId == sThreadId)
		{
			// This is a potential deadlock but ever detected in OnMutexTryLock
			// We should go here only under windows because a lock can be locked several times
			// by the same thread (without unlock)
		}
		else
		{
			bDeadlockFound = true;
			// This is a potential deadlock. The thread owning the lock has probably be
			// removed hardly without unlocking the lock explicitly
			sprintf(DeadlockBuffer, "*** Potential deadlock found for next mutex - A thread didnt' explicitly unlocked the mutex ***\n");
			WriteDeadlockFile();
			sprintf(DeadlockBuffer, "*** Origin mutex file and line is %s ***\n", sFileNameLine.c_str());
			WriteDeadlockFile();

		}
	}
	// We now set the owner of the mutex
	string sValue = sThreadId;
	sValue += ";";
	sValue += sSourceFile;
	sValue += ";";
	sValue += sSourceFileLine;
	ActualLocks[sMutex] = sValue;
	if(bDeadlockFound || bDisplayAll)
	{
		sprintf(DeadlockBuffer, "th:%s - mx:%s - %-15s - File = %-15s - Line = %s\n", sThreadId.c_str(), sMutex.c_str(), sAction.c_str(), sSourceFile.c_str(), sSourceFileLine.c_str());
		WriteDeadlockFile();
	}
	ProtectCheckDeadLocks.Unlock();
}
void OnMutexUnlock(void * Mutex, const char *SourceFile, int SourceFileLine)
{
	ProtectCheckDeadLocks.Lock();
	string sSourceFile = GetSourceFileName(SourceFile);
	string sSourceFileLine = GetSourceFileLine(SourceFileLine);
	string sThreadId = GetCurrentThreadIdentifier();
	string sMutex = GetMutexString(Mutex);
	string sAction = "UNLOCK";
	bool bDeadlockFound = false;
	if(ActualLocks.find(sMutex) != ActualLocks.end())
	{
		// The mutex is ever in the locked list
		// Try to get the thread that locked it
		string sLockingThreadId = ActualLocks[sMutex];
		string sFileNameLine = "";
		string::size_type index;
		index = sLockingThreadId.find(';');
		if(index != string::npos)
		{
			sFileNameLine = sLockingThreadId.substr(index+1);
			sLockingThreadId = sLockingThreadId.substr(0, index);
		}
		if(sLockingThreadId == sThreadId)
		{
			// The thread correctly released the mutex
			// So we remove it from the list
			ActualLocks.erase(sMutex);
		}
		else
		{
			bDeadlockFound = true;
			// The thread unlocks a lock but is not the owner of this lock
			sprintf(DeadlockBuffer, "*** Potential deadlock found for next mutex - The thread unlocks a lock but is not the owner of this lock ***\n");
			WriteDeadlockFile();
			sprintf(DeadlockBuffer, "*** Origin mutex file and line is %s ***\n", sFileNameLine.c_str());
			WriteDeadlockFile();
		}
	}
	else
	{
		bDeadlockFound = true;
		// The thread unlocks a lock but there is no owner for this lock (not locked)
		sprintf(DeadlockBuffer, "*** Potential deadlock found for next line - The thread unlocks a lock but there is no owner for this lock ***\n");
		WriteDeadlockFile();
	}
	if(bDeadlockFound || bDisplayAll)
	{
		sprintf(DeadlockBuffer, "th:%s - mx:%s - %-15s - File = %-15s - Line = %s\n", sThreadId.c_str(), sMutex.c_str(), sAction.c_str(), sSourceFile.c_str(), sSourceFileLine.c_str());
		WriteDeadlockFile();
	}
	ProtectCheckDeadLocks.Unlock();
}
void WriteDeadlockFile()
{
	if(!CheckDeadlocksInitialized)
	{
		unlink(sDeadlockFileName.c_str());
		CheckDeadlocksInitialized = true;
	}
	FILE *stream = Nvfopen(sDeadlockFileName.c_str(), "a+t");
	if(stream == NULL)
		return;
	fwrite(DeadlockBuffer, 1, strlen(DeadlockBuffer), stream);
	fclose(stream);	       
}
string GetSourceFileName(const char *SourceFile)
{
	string sSourceFile = SourceFile;
	string::size_type index;
	index = sSourceFile.rfind('/');
	if(index != string::npos)
		sSourceFile = sSourceFile.substr(index+1);
	index = sSourceFile.rfind('\\');
	if(index != string::npos)
		sSourceFile = sSourceFile.substr(index+1);
	index = sSourceFile.rfind('.');
	if(index != string::npos)
		sSourceFile = sSourceFile.substr(0, index);
	return sSourceFile;
}
string GetSourceFileLine(int FileLine)
{
	string sSourceFileLine = "";
	char Buffer[100];
	sprintf(Buffer, "%05d", FileLine);
	sSourceFileLine = Buffer;
	return sSourceFileLine;
}
string GetCurrentThreadIdentifier()
{
	string CurrentThreadId = "";
#ifdef WIN32
	void *ThreadHandle = (void *)GetCurrentThreadId();			// Thread handle for window
#else
	void *ThreadHandle = (void *) pthread_self();				// Thread handle for unix
#endif
	char Buffer[100];
	sprintf(Buffer, "%016lx", (unsigned long)ThreadHandle);
	CurrentThreadId = Buffer;
	return CurrentThreadId;
}
string GetMutexString(void * Mutex)
{
	string sMutext = "";
	char Buffer[100];
	sprintf(Buffer, "%016lx", (unsigned long)Mutex);
	sMutext = Buffer;
	return sMutext;
}
#endif

//**********************************************************************************
//**********************************************************************************
// opens a stream but making it not heritable from child processes
//**********************************************************************************
//**********************************************************************************
FILE *Nvfopen(const char *filename, const char *mode)
{
	FILE * ProviFile;
#ifdef WIN32
	DWORD AccessMode = 0;
	DWORD ShareMode = FILE_SHARE_DELETE | FILE_SHARE_READ |FILE_SHARE_WRITE;
	DWORD CreateMode = 0;
	DWORD Flags = FILE_ATTRIBUTE_NORMAL;
	int AppendMode = 0;
	int TextMode = 0;
	if(strchr(mode, 't'))
		TextMode = 1;
	if(!strcmp(mode, "r") || !strcmp(mode, "rb") || !strcmp(mode, "rt"))
	{
		AccessMode = GENERIC_READ;
		CreateMode = OPEN_EXISTING;
	}
	if(!strcmp(mode, "w") || !strcmp(mode, "wb") || !strcmp(mode, "wt"))
	{
		AccessMode = GENERIC_WRITE;
		CreateMode = CREATE_ALWAYS;
	}
	if(!strcmp(mode, "a") || !strcmp(mode, "ab") || !strcmp(mode, "at"))
	{
		AccessMode = GENERIC_WRITE;
		CreateMode = OPEN_ALWAYS;
		AppendMode = 1;
	}
	if(!strcmp(mode, "r+") || !strcmp(mode, "r+b") || !strcmp(mode, "r+t"))
	{
		AccessMode = GENERIC_READ | GENERIC_WRITE;
		CreateMode = OPEN_EXISTING;
	}
	if(!strcmp(mode, "w+") || !strcmp(mode, "w+b") || !strcmp(mode, "w+t"))
	{
		AccessMode = GENERIC_READ | GENERIC_WRITE;
		CreateMode = CREATE_ALWAYS;
	}
	if(!strcmp(mode, "a+") || !strcmp(mode, "a+b") || !strcmp(mode, "a+t"))
	{
		AccessMode = GENERIC_READ | GENERIC_WRITE;
		CreateMode = OPEN_ALWAYS;
		AppendMode = 1;
	}

	SECURITY_ATTRIBUTES sa;
	sa.nLength = sizeof(sa);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = false;

	// We try to open the file
	HANDLE hFile = CreateFile(filename, AccessMode, ShareMode, &sa, CreateMode, Flags, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		// if we are out of resources, we try again for 10 seconds
		DWORD LastError = GetLastError();
		if((LastError == ERROR_TOO_MANY_OPEN_FILES) ||
			(LastError == ERROR_NOT_ENOUGH_MEMORY) ||
			(LastError == ERROR_OUTOFMEMORY))
		{
			time_t initial_time = time(NULL);
			while(hFile == INVALID_HANDLE_VALUE)
			{
				SLEEP(500);	// we try every 500 milliseconds
				hFile = CreateFile(filename, AccessMode, ShareMode, &sa, CreateMode, Flags, NULL);
				if((int)(time(NULL) - initial_time) > 10)
					break;
			}
		}
	}
	if(hFile == INVALID_HANDLE_VALUE)
		return NULL;

	int flags = 0;
	if(!(AccessMode & GENERIC_WRITE))
		flags |= _O_RDONLY;
	if(AppendMode)
		flags |= _O_APPEND;
	if(TextMode)
		flags |= _O_TEXT;
	int CrtFileHandle = _open_osfhandle((intptr_t)hFile, flags); 
	
	if(CrtFileHandle == -1)
		return NULL;
	ProviFile = _fdopen(CrtFileHandle, mode);

#else
	ProviFile = fopen64(filename, mode);
	if(ProviFile == NULL)
	{
		// if we are out of resources, we try again for 10 seconds
		if((errno == EMFILE) ||
			(errno == ENFILE) ||
			(errno == ENOMEM))
		{
			time_t initial_time = time(NULL);
			while(ProviFile == NULL)
			{
				SLEEP(500);	// we try every 500 milliseconds
				ProviFile = fopen(filename, mode);
				if((int)(time(NULL) - initial_time) > 10)
					break;
			}
		}
	}
	if(ProviFile != NULL)
	{
		int CrtFileHandle = fileno(ProviFile);
		fcntl(CrtFileHandle, F_SETFD, FD_CLOEXEC);
	}
#endif

	return ProviFile;
}

NvInt64 Nvftell(FILE *File)
{
#ifdef WIN32
	return _ftelli64(File);
#else
	return (NvInt64)ftello64(File);
#endif
}

int Nvfseek(FILE *File, NvInt64 Offset, int Origin)
{
#ifdef WIN32
	return _fseeki64(File, Offset, Origin);
#else
	return fseeko64(File, Offset, Origin);
#endif
}


NvInt64 NvAtoi64(const char *str)
{
#ifdef WIN32
	return _atoi64(str);
#endif
#ifdef LINUX
	return atoll(str);
#endif
#ifdef AIX
	return atoll(str);
#endif
#ifdef SOLARIS
	return atoll(str);
#endif
#ifdef HPUX
	return strtoimax(str, (char **)NULL, 10);
#endif
#ifdef HPUXI
	return strtoimax(str, (char **)NULL, 10);
#endif
}

// NvMutex class

// constructor
NvMutex::NvMutex()
{
#ifdef WIN32
	m_mMutex = CreateMutex(NULL, 0, NULL);
#else
	pthread_mutex_init(&m_mMutex, NULL);
#endif
}


// constructor by copy
// we don't accept to copy a mutex, we create another one
NvMutex::NvMutex(const NvMutex& MutexSrc)
{
#ifdef WIN32
	m_mMutex = CreateMutex(NULL, 0, NULL);
#else
	pthread_mutex_init(&m_mMutex, NULL);
#endif
}


// destructor
NvMutex::~NvMutex()
{
#ifdef WIN32
	ReleaseMutex(m_mMutex);
	CloseHandle(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
	pthread_mutex_destroy(&m_mMutex);
#endif
}


// lock
void NvMutex::Lock()
{
#ifdef WIN32
	WaitForSingleObject(m_mMutex, INFINITE);
#else
	pthread_mutex_lock(&m_mMutex);
#endif
}


// unlock
void NvMutex::Unlock()
{
#ifdef WIN32
	ReleaseMutex(m_mMutex);
#else
	pthread_mutex_unlock(&m_mMutex);
#endif
}

// Assignement operator
const NvMutex& NvMutex::operator = (const NvMutex& MutexSrc)
{
	return *this;	// we do nothing
}




// NvEvent class

// constructor
NvEvent::NvEvent()
{
	m_iValue = 0;
	m_bAutoReset = false;
	m_bSignaled = false;
	m_bAfterLock = false;
#ifdef WIN32
	m_eEvent = CreateEvent(NULL, 1, 0, NULL);
#else
	pthread_cond_init(&m_eEvent, 0);
#endif
}


// constructor
NvEvent::NvEvent(bool bAutoReset)
{
	m_iValue = 0;
	m_bAutoReset = bAutoReset;
	m_bSignaled = false;
	m_bAfterLock = false;
#ifdef WIN32
	m_eEvent = CreateEvent(NULL, 1, 0, NULL);
#else
	pthread_cond_init(&m_eEvent, 0);
#endif
}


// constructor by copy
// we don't accept to copy an event, we create another one
// we just copy the auto reset mode
NvEvent::NvEvent(const NvEvent& EventSrc)
{
	m_bAutoReset = EventSrc.m_bAutoReset;
	m_iValue = 0;
	m_bSignaled = false;
	m_bAfterLock = false;
#ifdef WIN32
	m_eEvent = CreateEvent(NULL, 1, 0, NULL);
#else
	pthread_cond_init(&m_eEvent, 0);
#endif
}


// destructor
NvEvent::~NvEvent()
{
	if(!m_bAfterLock)	// this test should work without mutext protection
		m_mMutex.Lock();	// we first lock the mutex for being sure nobody is using it
#ifdef WIN32
	ResetEvent(m_eEvent);
	CloseHandle(m_eEvent);
#else
	pthread_cond_destroy(&m_eEvent);
#endif
}


// lock
int NvEvent::Lock(int iTimeOut)
{
	m_mMutex.Lock();	// we first lock
	
	int iRet = NV_RET_OK;		// the return value is 1 (OK) by default

	if(m_bSignaled)				// if signaled, it is not necessary to wait
	{
		m_bAfterLock = true;	// we stay locked
		return iRet;
	}

	if(iTimeOut == NV_PARAM_NOTIMEOUT)
		return NV_RET_RESET;

	if(iTimeOut == 0)
		iTimeOut = 1;


#ifdef WIN32
	m_mMutex.Unlock();		// we unlock the mutex
	
	// Now we are waiting for the event to be signaled
	DWORD dwRet;
	if(iTimeOut < 0)
		dwRet = WaitForSingleObject(m_eEvent, INFINITE);
	else
		dwRet = WaitForSingleObject(m_eEvent, iTimeOut);

	m_mMutex.Lock();		// we lock again the mutex
	
	if(dwRet != WAIT_OBJECT_0)
		iRet = NV_RET_ERROR;

	if(dwRet == WAIT_TIMEOUT)
		iRet = NV_RET_TIMEOUT;
#else
	// under unix, the wait function directly unlock the mutex and relocks it after condition occurs or time out
	if(iTimeOut <= 0)
		pthread_cond_wait(&m_eEvent, &m_mMutex.m_mMutex);
	else
	{
		struct timeval now;
		struct timezone tz;
		struct timespec timeout;
		gettimeofday(&now, &tz);
		timeout.tv_sec= now.tv_sec + (iTimeOut/1000);
		timeout.tv_nsec= (now.tv_usec + ((iTimeOut%1000)*1000))*1000;
		if(timeout.tv_nsec > 1000000000)
		{
			timeout.tv_nsec -= 1000000000;
			timeout.tv_sec += 1;
		}
		int iRetFunction = pthread_cond_timedwait(&m_eEvent, &m_mMutex.m_mMutex, &timeout);
		if(iRetFunction != 0)
			iRet = NV_RET_ERROR;

		if(iRetFunction == ETIMEDOUT)
			iRet = NV_RET_TIMEOUT;
	}
#endif
	if(!m_bSignaled)				// Maybe the event has been reset by another thread
		iRet = NV_RET_RESET;

	m_bAfterLock = true;	// we stay locked
	return iRet;
}


// unlock
void NvEvent::Unlock()
{
	if(m_bAutoReset)
	{
		m_bSignaled = false;
#ifdef WIN32
		ResetEvent(m_eEvent);
#endif
	}

	// we are not locked now
	m_bAfterLock = false;
	m_mMutex.Unlock();
}

// signal
void NvEvent::Signal(int iValue)
{
	m_mMutex.Lock();	// we first lock
	m_iValue = iValue;	// set the value
	
	// now we can signal
	m_bSignaled = true;
#ifdef WIN32
	SetEvent(m_eEvent);
#else
	pthread_cond_broadcast(&m_eEvent);
#endif
	m_mMutex.Unlock();	// we unlock
}


// Release
// Can be be used only outside of Lock / Unlock block
void NvEvent::Reset()
{

	m_mMutex.Lock();	// we first lock

	ResetInside();
	
	m_mMutex.Unlock();	// we unlock
}


// ReleaseInside
// Can be used only inside a Lock / Unlock block
void NvEvent::ResetInside()
{
	// now we are noy signaled any more
	m_bSignaled = false;

#ifdef WIN32
	ResetEvent(m_eEvent);
#endif
}



// IsSignaled
// Can be used only outside a Lock / Unlock block
bool NvEvent::IsSignaled()
{
	m_mMutex.Lock();	// we first lock
	
	bool bRet = m_bSignaled;
	m_mMutex.Unlock();	// we unlock

	return bRet;
}


// IsSignaledInside
// Can be used only inside a Lock / Unlock block
bool NvEvent::IsSignaledInside()
{
	return m_bSignaled;
}


// GetValue
// Can be used only inside a Lock / Unlock block
int NvEvent::GetValue()
{
	return m_iValue;
}


// Assignement operator
const NvEvent& NvEvent::operator = (const NvEvent& EventSrc)
{
	m_bAutoReset = EventSrc.m_bAutoReset;
	return *this;	// we do nothing else
}




// NvSemaphore class

// constructor
NvSemaphore::NvSemaphore()
{
#ifdef WIN32
	s_sSemaphore = CreateSemaphore(NULL, 1, 1000000, NULL);	// we must set a maximum value
#else
#ifdef AIX
	m_iValue = 1;	// the initial value is 1 by default
#else
	sem_init(&s_sSemaphore, 0, 1);
#endif
#endif
}


// constructor with initial value
NvSemaphore::NvSemaphore(int iValue)
{
#ifdef WIN32
	s_sSemaphore = CreateSemaphore(NULL, iValue, 1000000, NULL);	// we must set a maximum value
#else
#ifdef AIX
	m_iValue = iValue;
#else
	sem_init(&s_sSemaphore, 0, iValue);
#endif
#endif
}


// constructor by copy
// we don't accept to copy a semaphore, we create another one with value 1
NvSemaphore::NvSemaphore(const NvSemaphore& SemaphoreSrc)
{
#ifdef WIN32
	s_sSemaphore = CreateSemaphore(NULL, 1, 1000000, NULL);	// we must set a maximum value
#else
#ifdef AIX
	m_iValue = 1;	// the initial value is 1 by default
#else
	sem_init(&s_sSemaphore, 0, 1);
#endif
#endif
}


// destructor
NvSemaphore::~NvSemaphore()
{
#ifdef WIN32
	CloseHandle(s_sSemaphore);
#else
#ifndef AIX
	sem_destroy(&s_sSemaphore);
#endif
#endif
}


// lock
void NvSemaphore::Lock()
{
#ifdef WIN32
	WaitForSingleObject(s_sSemaphore, INFINITE);
#else
#ifdef AIX
	m_Event.Lock();				// the event is signaled
								// only if the Value is not 0
	m_Mutex.Lock();				// m_iValue access is protected
	m_iValue--;					// we decrease the semaphore value
	if(m_iValue == 0)			// the event must be locked
		m_Event.ResetInside();	// if value is 0
	m_Mutex.Unlock();			// we unlock the mutex
	m_Event.Unlock();			// now, we unlock the event
#else
	sem_wait(&s_sSemaphore);
#endif
#endif
}


// trylock
bool NvSemaphore::TryLock()
{
#ifdef WIN32
	if(WaitForSingleObject(s_sSemaphore, 1)  != WAIT_OBJECT_0)
		return false;
#else
#ifdef AIX
	m_Mutex.Lock();				// m_iValue access is protected
	if(m_iValue == 0)
	{
		m_Mutex.Unlock();
		return false;
	}
	m_iValue--;					// we decrease the semaphore value
	if(m_iValue == 0)			// the event must be locked
		m_Event.Reset();		// if value is 0
	m_Mutex.Unlock();
#else
	if(sem_trywait(&s_sSemaphore) != 0)
		return false;
#endif
#endif
	return true;
}


// unlock
void NvSemaphore::Unlock()
{
#ifdef WIN32
	ReleaseSemaphore(s_sSemaphore, 1, NULL);
#else
#ifdef AIX
	m_Mutex.Lock();				// m_iValue access is protected
	m_iValue++;					// we increase the value
	m_Event.Signal();			// and signal the event
	m_Mutex.Unlock();
#else
	sem_post(&s_sSemaphore);
#endif
#endif
}


// set
void NvSemaphore::Set(int iValue)
{
	if(iValue <=0)		// we must have a positive value
		return;

	while(TryLock());	// we reset it to 0

	for(int i=0; i<iValue; i++)	// we increase it until iValue
		Unlock();
}


// get
int NvSemaphore::Get()
{
#ifdef WIN32
	LONG iValue;
	ReleaseSemaphore(s_sSemaphore, 1, &iValue);	// we increase by one
	Decrease(1);
	return (int)iValue;
#else
	int iValue;
#ifdef AIX
	m_Mutex.Lock();				// m_iValue access is protected
	iValue = m_iValue;
	m_Mutex.Unlock();
#else
	sem_getvalue(&s_sSemaphore, &iValue);
#endif
	return iValue;
#endif
}


// decrease
void NvSemaphore::Decrease(int iValue)
{
	if(iValue <=0)		// we must have a positive value
		return;

	for(int i=0; i<iValue; i++)
	{
		if(!TryLock())	// we decrease but stop at 0;
			break;
	}
}


// increase
void NvSemaphore::Increase(int iValue)
{
	if(iValue <=0)		// we must have a positive value
		return;
	for(int i=0; i<iValue; i++)	// we increase it until iValue
		Unlock();
}


// Assignement operator
const NvSemaphore& NvSemaphore::operator = (const NvSemaphore& SemaphoreSrc)
{
	return *this;	// we do nothing else
}



/*

// NvLog class

// constructor
NvLog::NvLog()
{
	m_sLogFile = "";			// no log file by default
	m_bEnableDebug = false;		// Log is disabled
}


// constructor with initial value
NvLog::NvLog(bool bEnableDebug)
{
	m_sLogFile = "";
	m_bEnableDebug = bEnableDebug;
}


// Enable log
void NvLog::EnableLog(bool bEnableDebug)
{
	Lock();
	m_bEnableDebug = bEnableDebug;
	Unlock();
}


// writing to Log file
void NvLog::WriteMessage(string sMessage)
{
	// we return if nothing to write
	if(!sMessage.length())
		return;

	Lock();	// we must lock
	if(!m_bEnableDebug)
	{
		Unlock();
		return;
	}
	if(!m_sLogFile.length())
	{
		Unlock();
		return;
	}
	
	FILE * File = Nvfopen(m_sLogFile.c_str(), "ab");
	if(File != NULL)
	{
		char szut[50];
#ifdef WIN32
		sprintf(szut, "%016x - ", GetCurrentThreadId());
#else
		sprintf(szut, "%016x - ", pthread_self());
#endif
		// get current time and date
		time_t CurrentTime;
		time(&CurrentTime);
		struct tm *tCurrentTime = localtime(&CurrentTime);
		char sTime[30];
		
		sprintf(sTime, "%02d/%02d/%4d - %02d:%02d:%02d - ",
						tCurrentTime->tm_mday,
						tCurrentTime->tm_mon + 1,
						tCurrentTime->tm_year + 1900,
						tCurrentTime->tm_hour,
						tCurrentTime->tm_min,
						tCurrentTime->tm_sec);
		fwrite(szut, 1, strlen(szut), File);
		fwrite(sTime, 1, strlen(sTime), File);
		fwrite(sMessage.c_str(), 1, sMessage.length(), File);
#ifdef WIN32
		strcpy(szut, "\r\n");
#else
		strcpy(szut, "\n");
#endif
		fwrite(szut, 1, strlen(szut), File);
		fclose(File);
	}

	
	Unlock();
}


// Init the log file
void NvLog::Init(string sLogFile)
{
	Lock();
	m_sLogFile = sLogFile;
	if(!m_sLogFile.length())
	{
		Unlock();
		return;
	}
	
	// reset the log file to a 0 length
	FILE * File = Nvfopen(m_sLogFile.c_str(), "wb");
	if(File != NULL)
		fclose(File);
	Unlock();
}

*/


// Class NvTask
// see comments in header file

// constructor
NvTask::NvTask()
{
	m_bSameThread = false;
	m_bAutoDelete = false;
	m_bThreadRunning = false;
#ifdef WIN32
	m_ThreadHandle = NULL;
#else
	m_ThreadHandle = (pthread_t)NULL;
#endif

#ifndef WIN32
	NV_LOCKLOCK(&m_mProtectPrivate);
	KillRequested = false;
	NV_UNLOCKLOCK(&m_mProtectPrivate);
#endif
}


// destructor
NvTask::~NvTask()
{
	// first sets the autodelete to false
	// because we are in the destructor
	NV_LOCKLOCK(&m_mProtectPrivate);
	m_bAutoDelete = false;
	NV_UNLOCKLOCK(&m_mProtectPrivate);

	// Now kill the thread if not done
	// The kill function also calls the OnCleanup
	Kill();
}



// Thread procedure (static function)
#ifdef WIN32
void __cdecl NvTask::ThreadMain(void *Value)
#else
void* NvTask::ThreadMain(void *Value)
#endif
{
	// As The thread procedure can only be declared as a static function
	// We pass the this pointer as parameter.
	// Then we call the Main function of the this pointer
	// The Main function is a pure virtual function that must be overloaded
	// in derived class.

	NvTask *Task = (NvTask *)Value;
	
#ifndef WIN32
	
	// We block terminaison signals for this thread.
	// Because the termination is managed at the process level
	sigset_t newmask;
	sigemptyset(&newmask);
	sigaddset(&newmask, SIGINT);
	sigaddset(&newmask, SIGTERM);
#ifdef AIX
	sigaddset(&newmask, SIGQUIT);
	sigaddset(&newmask, SIGIOT);
	sigaddset(&newmask, SIGTSTP);
#endif
#ifdef AIX
	sigthreadmask(SIG_BLOCK, &newmask, NULL);
#else
	pthread_sigmask(SIG_BLOCK, &newmask, NULL);
#endif

	// we install cleanup function for scratch under unix
	// the pair push and pop must be in the same function
	// that's why we don't implementing it in
	// OnInit and OnCleanup
	pthread_cleanup_push(NvTask::CleanUpUnixThread, Value);
#endif

	// init the thread
	Task->m_iExitCode = 1;
	if(Task->OnInit())
	{
		// call the Main function and get its error code
		try
		{
			Task->m_iExitCode = Task->Main();
		}
		catch (exception & ex)
		{
			Task->OnException(ex.what());
		}
		catch(...)
		{
			Task->OnException("Unknown");
		}
	}

#ifndef WIN32
	pthread_cleanup_pop(0);
#endif

	// clean the thread
	Task->OnCleanup();

#ifdef WIN32
	return;						// end of thread (windows)
#else
	pthread_detach(pthread_self());
	pthread_exit(NULL);

	return (void *)(0);			// end of thread (unix)
#endif
}


#ifndef WIN32
// Cleanup thread procedure for unix (static function)

void NvTask::CleanUpUnixThread(void *Value)
{
	// Value contains a pointer to a NvTask class (or derived class)

	NvTask *Task = (NvTask *)Value;

	// clean the thread
	NV_LOCKLOCK(&(Task->m_mProtectPrivate));
	// If a kill order has been send, its not considered as an exception
	if(Task->KillRequested)
	{
		Task->KillRequested = false;
		NV_UNLOCKLOCK(&(Task->m_mProtectPrivate));
	}
	else
	{
		NV_UNLOCKLOCK(&(Task->m_mProtectPrivate));
		Task->OnException("Uncatched exception (signal occured)");
	}
	Task->OnCleanup();
}
#endif


// Called for thread initialization
bool NvTask::OnInit()
{
	NV_LOCKLOCK(&m_mProtectPrivate);
	// set the thread handle
#ifdef WIN32
	HANDLE SourceHandle = GetCurrentThread();
	HANDLE DestinationHandle = 0;
	DuplicateHandle(GetCurrentProcess(), SourceHandle, GetCurrentProcess(), &DestinationHandle, 0, TRUE, DUPLICATE_SAME_ACCESS);
	m_ThreadHandle = DestinationHandle;
#else
	m_ThreadHandle = pthread_self();
#endif
	NV_UNLOCKLOCK(&m_mProtectPrivate);

	return true;
}



// Called after main terminates
void NvTask::OnCleanup()
{

	NV_LOCKLOCK(&m_mProtectPrivate);

	// The thread handle is not valid now
#ifdef WIN32
	if(m_ThreadHandle != NULL)
		CloseHandle(m_ThreadHandle);
	m_ThreadHandle = NULL;
#else
	m_ThreadHandle = (pthread_t)NULL;
#endif
	
	// Now the thread is not running any more
	m_bThreadRunning = false;
	
	// we destroy the object if the auto delete mode has been requested
	if(m_bAutoDelete)
	{
		NV_UNLOCKLOCK(&m_mProtectPrivate);
		delete this;
	}
	else
		NV_UNLOCKLOCK(&m_mProtectPrivate);

}


// test if thread is running
bool NvTask::IsThreadRunning()
{
	// We test if the m_bThreadRunning variable is true
	// We also manage the possiblity if this function to be
	// called after the object has been deleted
	// It is forbiden to call a function of a deleted object
	// but it may occur
	bool bThreadRunning;
	try
	{
		NV_LOCKLOCK(&m_mProtectPrivate);
		bThreadRunning = m_bThreadRunning;
		NV_UNLOCKLOCK(&m_mProtectPrivate);
	}
	catch(...)
	{
		return false;
	}
	
	if(bThreadRunning != true)
		return false;
	return true;
}


// called when an exception occurs
void NvTask::OnException(string sInfo)
{
	// this function does nothing by default
	// it may be overloaded by derived classes
}


// starts the thread if requested
bool NvTask::Run(bool bSameThread, bool bAutoDelete)
{
	// we accept to change the autodelete mode in any case
	NV_LOCKLOCK(&m_mProtectPrivate);
	m_bAutoDelete = bAutoDelete;
	NV_UNLOCKLOCK(&m_mProtectPrivate);

	if(IsThreadRunning())
	{
		if(bSameThread)		// we cannot change that while the thread is running
			return false;
		return true;		// the thread is ever running
	}

	
	// we change the samethread flag
	NV_LOCKLOCK(&m_mProtectPrivate);
	m_bSameThread = bSameThread;
	NV_UNLOCKLOCK(&m_mProtectPrivate);

	// The exit code is 1 by default
	m_iExitCode = 1;

	// if the same thread is requested, we just call the Main function
	// and deletes the object if auto delete flag is set
	if(bSameThread)
	{
		// Call the init function
		bool RetVal = true;
		if(OnInit())
		{
			// Call the main function
			try
			{
				m_iExitCode = Main();
			}
			catch (exception & ex)
			{
				OnException(ex.what());
			}
			catch(...)
			{
				OnException("Unknown");
			}
		}
		else
			RetVal = false;
		// Call the cleanup function
		OnCleanup();
		return RetVal;
	}


	// A separate thread has been asked
	// So we start it

	// We suppose that the thread is running
	NV_LOCKLOCK(&m_mProtectPrivate);
	m_bThreadRunning = true;
	NV_UNLOCKLOCK(&m_mProtectPrivate);
#ifdef WIN32
		if(_beginthread(NvTask::ThreadMain, 0, this) == -1)
		{
			NV_LOCKLOCK(&m_mProtectPrivate);
			m_bThreadRunning = false;	// no the thread is not running
			NV_UNLOCKLOCK(&m_mProtectPrivate);
			return false;
		}
#else
		pthread_attr_t	attr;
		if(pthread_attr_init(&attr) == -1)
		{
			NV_LOCKLOCK(&m_mProtectPrivate);
			m_bThreadRunning = false;	// no the thread is not running
			NV_UNLOCKLOCK(&m_mProtectPrivate);
			return false;
		}
		
		pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);		// the unix thread has a system scope for best performances
		pthread_t thid;
		if(pthread_create(&thid, &attr, NvTask::ThreadMain, this) == -1)
		{
			NV_LOCKLOCK(&m_mProtectPrivate);
			m_bThreadRunning = false;	// no the thread is not running
			NV_UNLOCKLOCK(&m_mProtectPrivate);
			return false;
		}
		pthread_attr_destroy(&attr);
#endif

	return true;
}


// this hardly kills a thread
void NvTask::Kill()
{
	// do nothing if no thread is running
	if(!IsThreadRunning())
		return;
	NV_LOCKLOCK(&m_mProtectPrivate);
#ifdef WIN32
	if(m_ThreadHandle != NULL)	// terminate the thread
	{
		NV_UNLOCKLOCK(&m_mProtectPrivate);
		TerminateThread(m_ThreadHandle, -1);
		SLEEP(200);	// let time for thread to terminate
	}
	else
		NV_UNLOCKLOCK(&m_mProtectPrivate);
	// we call the cleanup function
	OnCleanup();
#else
	if(m_ThreadHandle)
	{
		// The kill command will automatically generate an exception
		// that will call the OnCleanup function
		KillRequested = true;
		NV_UNLOCKLOCK(&m_mProtectPrivate);
		pthread_kill(m_ThreadHandle, SIGSEGV);
		SLEEP(200);	// let time for thread to terminate
	}
	else
		NV_UNLOCKLOCK(&m_mProtectPrivate);
#endif
}





// Class NvMsg
// see comments in header file
NvMsg::NvMsg()
{
	m_cProcessTime = 0;
	m_bProcessed = false;
	m_bUseTimeFunction = true;
	m_iTimeOut = 300;		// 5 minutes by default
	m_iPriority = 0;
}

NvMsg::NvMsg(int TimeOut)
{
	m_cProcessTime = 0;
	m_bProcessed = false;
	m_bUseTimeFunction = true;
	m_iTimeOut = TimeOut;
	m_iPriority = 0;
}

NvMsg::NvMsg(int TimeOut, int Priority)
{
	m_cProcessTime = 0;
	m_bProcessed = false;
	m_bUseTimeFunction = true;
	m_iTimeOut = TimeOut;
	m_iPriority = Priority;
}

NvMsg::~NvMsg()
{
	// Nothing special to do
}



// The overload of this operator allows priority comparaison
bool NvMsg::operator < (const NvMsg& NvMsgSrc) const
{
	// compare priority values
	return (m_iPriority < NvMsgSrc.m_iPriority);
}

	
	
	
// Returns the Time required for processing the message in milliseconds
// This can be called after retrieving the message from the NvMessageTask
// Output queue
int NvMsg::GetProcessTime()
{
	if(!m_bProcessed)				// The process time is 0 if the message has not been processed
		return 0;

	if(!m_bUseTimeFunction)		// The process time is 0 if we don't use the time functions
		return 0;
	
	return (int)(m_cProcessTime * (clock_t)1000)/CLOCKS_PER_SEC ;	// otherwise, we return the process time
}


// Process function called by the NvMessageTask MainLoop function
// The derived classes should overload this function but have to call
// the base class before doing anything else
void NvMsg::Process()
{
	if(m_bProcessed)
		return;

	if(m_bUseTimeFunction)
		m_cStartProcessTime = clock();
}


// Process function called by the NvMessageTask MainLoop function
// The derived classes should overload this function but have to call
// the base class before doing anything else
void NvMsg::EndProcess()
{
	m_bProcessed = true;

	if(m_bUseTimeFunction)
	{
		m_cProcessTime = clock();
		m_cProcessTime = m_cProcessTime - m_cStartProcessTime;
		m_cStartProcessTime = clock();
	}
}


// IsExpired function called by the NvMessageTask CheckTimeOuts function
bool NvMsg::IsExpired()
{
	if(m_bProcessed)
		return false;

	if(!m_bUseTimeFunction)
		return false;

	clock_t	cCloclTick;
	cCloclTick = CLOCKS_PER_SEC;
	clock_t	cTimeFromProcess = ((clock() - m_cStartProcessTime)) / cCloclTick;
	if(cTimeFromProcess <= m_iTimeOut)
		return false;

	return true;
}

#ifndef WIN32

// Init signals than must be caugth at thread level
// This function must be called only one time from the main
// This is a static function
void InitThreadSignalHandler()
{
	NV_LOCKLOCK(&ProtectUnixSignalFunc);
	// we do the following only if required
	// the thread is catching all these signals
	signal(SIGABRT, CatchSignalThread);
	signal(SIGFPE, CatchSignalThread);
	signal(SIGILL, CatchSignalThread);
	signal(SIGSEGV, CatchSignalThread);

	signal(SIGHUP, CatchSignalThread);
	signal(SIGTRAP, CatchSignalThread);
#ifdef AIX
	signal(SIGEMT, CatchSignalThread);
	signal(SIGSYS, CatchSignalThread);
	signal(SIGPRE, CatchSignalThread);
#endif
	signal(SIGBUS, CatchSignalThread);
//		signal(SIGPIPE, CatchSignalThread);
	signal(SIGTTIN, CatchSignalThread);
	signal(SIGTTOU, CatchSignalThread);
	signal(SIGXCPU, CatchSignalThread);
	signal(SIGXFSZ, CatchSignalThread);
	signal(SIGCHLD, CatchSignalThread);

	NV_UNLOCKLOCK(&ProtectUnixSignalFunc);
}

// this function is called when a signal is sent to the thread
void CatchSignalThread(int SigNum)
{
	if(SigNum == SIGCHLD)
		return;
	if(SigNum == SIGABRT)
	{
		NV_LOCKLOCK(&ProtectUnixSignalFunc);
		signal(SigNum, CatchSignalThread);	// we must reset the singal handling function
		NV_UNLOCKLOCK(&ProtectUnixSignalFunc);
		pthread_detach(pthread_self());
		pthread_exit(NULL);
		return;
	}
printf("**************** Signal occured %d\n", SigNum);	
	NV_LOCKLOCK(&ProtectUnixSignalFunc);
	signal(SigNum, CatchSignalThread);	// we must reset the singal handling function
	NV_UNLOCKLOCK(&ProtectUnixSignalFunc);

	// We remove the thread under unix when such signal occurs
	// This will correctly cleanup the thread by calling the NvTsk::OnCleanup
	// because we have installed a handler for thread termination
	// Under linux there is a possibility to generate exceptions on signal
	// The code must be compiled with -fnon-call-exceptions
#ifdef LINUX
	sigset_t x;
	sigemptyset (&x);
	sigaddset(&x, SigNum);
	pthread_sigmask(SIG_UNBLOCK, &x, NULL);

	throw std::exception();
#else
	pthread_detach(pthread_self());
	pthread_exit(NULL);
#endif

}
#endif
